from qgis.PyQt.QtCore import (QCoreApplication, QT_TRANSLATE_NOOP)
from qgis.PyQt.QtGui import QColor
from qgis.core import (
  QgsCoordinateReferenceSystem,
  QgsProcessingContext,
  QgsProcessingFeedback,
  QgsProcessingParameterExtent,
  QgsProcessingParameterDistance,
  QgsProcessingParameterCrs, 
  QgsProcessingParameterRasterDestination,
  QgsProcessingParameterString,
  QgsRasterLayer,
  QgsProcessingParameterString
  )
from qgis import processing

from .fetchabstract import fetchabstract
from ..algpostprocessor import hrPostProcessor
from ..hr_config import PostProcessors

import math
import itertools

class fetchpopghs(fetchabstract):
  
  PARAMETERS = {  
    "FETCH_EXTENT": {
      "ui_func": QgsProcessingParameterExtent,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("fetchpopghs","Extent for fetching data")
      }
    },
    "TARGET_CRS": {
      "ui_func": QgsProcessingParameterCrs,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("fetchpopghs","Target CRS (Cartesian coordinates)")
      }
    },
    "BUFFER": {
      "ui_func": QgsProcessingParameterDistance,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("fetchpopghs","Buffer of the fetch area (using Target CRS)"),
        "defaultValue": 0.0,
        "parentParameterName": "TARGET_CRS"
      }
    },
    "WEBFETCH_URL": {
      "ui_func": QgsProcessingParameterString,
      "ui_args": {
        "optional": True,
        "description": QT_TRANSLATE_NOOP("fetchpopghs","Base-URL of the GHS data"),
        "defaultValue": "https://jeodpp.jrc.ec.europa.eu/ftp/jrc-opendata/GHSL/GHS_POP_GLOBE_R2023A/GHS_POP_E2020_GLOBE_R2023A_4326_3ss/V1-0/tiles/GHS_POP_E2020_GLOBE_R2023A_4326_3ss_V1_0_"
      }
    },
    "OUTPUT": {
      "ui_func": QgsProcessingParameterRasterDestination,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("fetchpopghs","Population" )
      }
    }
  }  
  
  
  def setWebFetchArgs(self, parameters, context, feedback):
    
    self.WEBFETCH_ARGS["REQUEST_ARGS"] = []
    
    # set url and file names
    lat_min = math.floor(self.FETCH_AREA.yMinimum()/10) * 10
    lat_max = math.floor(self.FETCH_AREA.yMaximum()/10) * 10 + 1
    lng_min = math.floor(self.FETCH_AREA.xMinimum()/10) * 10
    lng_max = math.floor(self.FETCH_AREA.xMaximum()/10) * 10 + 1
    
    base_url = self.parameterAsString(parameters, "WEBFETCH_URL", context)
    
    for lat, lng in itertools.product(list(range(lat_min, lat_max)), list(range(lng_min, lng_max))):
      lng_str = f"C{int((lng+180)/10) + 1}"
      lat_str = f"R{int((90-lat)/10)}"
      self.WEBFETCH_ARGS["REQUEST_ARGS"].append(
        {"method": "GET","url": base_url + f"{lat_str}_{lng_str}.zip"}
      )

    if self.WEBFETCH_ARGS["REQUEST_ARGS"] is not None:
      self.WEBFETCH_ARGS["SET"] = True
    
  
  
  # create the raster from downloaded hgt files
  def fetchFeaturesFromWeb(self, parameters: dict, context: QgsProcessingContext, feedback: QgsProcessingFeedback, parallel = False) -> None:
    self.downloadFilesConcurrently(parameters, context, feedback, parallel=parallel)
    
    # unpack zip files
    files_to_be_merged = list(
      itertools.chain.from_iterable(
        map(
          lambda x: self.extractArchive(x, feedback, extension=".zip", pattern = "\\.tif$"), 
          self.WEBFETCH_ARGS["DOWNLOADED_FILE"]
        )
      )
    )
    
    # if there are no files
    if len(files_to_be_merged) == 0:
      feedback.reportError(self.tr("No pop files were downloaded."))
      raise Exception(self.tr("No pop files were downloaded."))
    
    # if there are only a file
    elif len(files_to_be_merged) == 1:
      self.FETCH_FEATURE = QgsRasterLayer(files_to_be_merged[0], "POP")
    
    # if there are multiple files
    else:
      rasters = []
      for file in files_to_be_merged:
        rasters.append(QgsRasterLayer(file + "|layername=1", "POP"))
      self.FETCH_FEATURE = processing.run(
        "gdal:merge",{
          "INPUT": rasters,
          "OUTPUT": "TEMPORARY_OUTPUT"
        },
        context = context,
        is_child_algorithm = True
      )["OUTPUT"]
    
  # initialize of the algorithm  
  def initAlgorithm(self, config):    
    self.initUsingCanvas()
    self.initParameters()

  # execution of the algorithm
  def processAlgorithm(self, parameters, context, feedback):   
    # set the fetch area, of which CRS is the same as SRTM's
    self.setFetchArea(parameters,context,feedback,QgsCoordinateReferenceSystem("EPSG:4326"))
    
    # check whether the CRS is long-lat coordinates
    if self.FETCH_AREA.crs().isGeographic() is not True:
      feedback.reportError(self.tr("The CRS is NOT a Geographic Coordinate System."))
      raise Exception(self.tr("The CRS is NOT a Geographic Coordinate System."))
      
    if self.FETCH_AREA.yMinimum() < -60 : 
      feedback.reportError(self.tr("The extent is out of GHS-covered area."))
      raise Exception(self.tr("The extent is out of GHS-covered area."))
    
    # set the meta information for obtaining SRTM data
    self.setWebFetchArgs(parameters, context, feedback)
    
    self.setCurrentProcess(use_caller=False)
    
    # download files using the session info
    self.fetchFeaturesFromWeb(parameters, context, feedback)
    
    # clip the raster because it is too large
    pop_raster_clipped = processing.run(
      "gdal:cliprasterbyextent", 
      {
        "INPUT": self.FETCH_FEATURE,
        "PROJWIN": self.FETCH_AREA,
        "OUTPUT": self.parameterAsOutputLayer(parameters, "OUTPUT", context)
      },
      context = context,
      is_child_algorithm = True
    )["OUTPUT"]
    
    # register history and postprocessor for the raster output
    self.setCurrentProcess(use_caller=False)
    self.registerPostProcessor(
      pop_raster_clipped,
      hrPostProcessor(history=[self.CURRENT_PROCESS], color_args = {"coloring": "single_band_pseudo_color", "theme": "Purples", "opacity": 0.5})
    )
    PostProcessors[pop_raster_clipped].setMinValToZero()
    
    return {"OUTPUT": pop_raster_clipped}   
    

  def displayName(self):
    return self.tr("Population (GHS)")

  def group(self):
    return self.tr('Fetch geometries')

  def groupId(self):
    return 'fetchgeometry'

  def createInstance(self):
    return fetchpopghs()

