from qgis.PyQt.QtCore import (QCoreApplication, QT_TRANSLATE_NOOP, QVariant)
from qgis.core import (
  QgsProcessingContext,
  QgsProcessingFeedback,
  QgsProcessingParameterCrs,
  QgsVectorLayer,
  QgsProcessingParameterExtent,
  QgsProcessingParameterDistance,
  QgsFeature,
  QgsPoint,
  QgsFeature,
  QgsProcessingParameterRasterDestination,
  QgsProcessingParameterString,
  QgsCoordinateTransform,
  QgsRectangle,
  QgsReferencedRectangle,
  QgsProject
  )
from qgis import processing

import itertools
import urllib
import json
import math
import os
from .fetchabstract import fetchabstract
from ..algpostprocessor import hrPostProcessor
from ..hr_config import PostProcessors

from ..worldmesh import (
  cal_meshcode5,
  meshcode_to_latlong
)

class fetchpopja(fetchabstract):
  
  PARAMETERS = {  
    "FETCH_EXTENT": {
      "ui_func": QgsProcessingParameterExtent,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("fetchpopja","Extent for fetching data")
      }
    },
    "TARGET_CRS": {
      "ui_func": QgsProcessingParameterCrs,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("fetchpopja","Target CRS (Cartesian coordinates)")
      }
    },
    "BUFFER": {
      "ui_func": QgsProcessingParameterDistance,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("fetchpopja","Buffer of the fetch area (using Target CRS)"),
        "defaultValue": 0.0,
        "parentParameterName": "TARGET_CRS"
      }
    },
    "WEBFETCH_URL": {
      "ui_func": QgsProcessingParameterString,
      "ui_args": {
        "optional": True,
        "description": QT_TRANSLATE_NOOP("fetchpopja","Base-URL of the population data"),
        "defaultValue": "https://jstatmap.e-stat.go.jp/statmap/api/1.00"
      }
    },
    "FETCH_CRS": {
      "ui_func": QgsProcessingParameterCrs,
      "ui_args": {
        "optional": True,
        "description": QT_TRANSLATE_NOOP("fetchpopja","CRS of the population data"),
        "defaultValue": "EPSG:6668"  # must be specified as string, because optional parameter cannot be set as QgsCoordinateReferenceSystem
      }
    },
    "OUTPUT": {
      "ui_func": QgsProcessingParameterRasterDestination,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("fetchpopja","Population" )
      }
    }
  }  
  
  
  ESTAT_ID_MESH_FILE = os.path.join(os.path.dirname(__file__),"estatId_mesh_list.txt")
  ESTAT_ID_MESH_DICT = {}
  
  ESTAT_API_URL = "https://jstatmap.e-stat.go.jp/statmap/api/1.00"   
  ESTAT_WHAT = {
    "STATISTICS_NAME": "令和2年国勢調査",
    "STAT_KIND": "人口",
    "TITLE": "TABLE01 男女別人口"
  }
    
  ESTAT_API_PARAMS = {
    "category": "richReport",
    "func": "getSummary",
    "key": "idfjPdvCUdVjhyamGgjm",
    "userid": "j.tagusari@eng.hokudai.ac.jp",
    "lat": 0.0,
    "lng": 0.0,
    "radius": 0.0,
    "output": "json"
  }
  
  
  LAT_UNIT_5 =  7.5 / 3600.0
  LNG_UNIT_5 = 11.25 / 3600.0
    
  # set information about the map tile
  def setWebFetchArgs(self, parameters, context, feedback):
        
    self.WEBFETCH_ARGS["CRS"] = self.parameterAsCrs(parameters, "FETCH_CRS", context)
    self.WEBFETCH_ARGS["GEOM_TYPE"] = "Point"
    self.WEBFETCH_ARGS["REQUEST_ARGS"] = []
    
    if not self.FETCH_AREA.crs().isGeographic():
      feedback.reportError(self.tr("The CRS of the fetch area must be geographic."))
      raise Exception(self.tr("The CRS of the fetch area must be geographic."))
    
    
    def cmptRadius(results, target_rect):
      xy_crs = self.getUtmCrs(lng=target_rect.center().x(), lat=target_rect.center().y())
      transform_to_xy = QgsCoordinateTransform(target_rect.crs(), xy_crs, QgsProject.instance())
      target_rect_xy = QgsReferencedRectangle(transform_to_xy.transformBoundingBox(target_rect), xy_crs)
      
      height = target_rect_xy.height()
      width = target_rect_xy.width()
      
      r = math.sqrt(height**2 + width**2) / 2 
      if r < 9000:
        results.append(
          {
            "lng": target_rect.center().x(), 
            "lat": target_rect.center().y(), 
            "radius": r
          }
        )
      else:
        lng_4 = [target_rect.xMinimum(), target_rect.center().x(), target_rect.xMaximum()]
        lat_4 = [target_rect.yMinimum(), target_rect.center().y(), target_rect.yMaximum()]
        rects = [
          QgsRectangle(lng_4[0], lat_4[0], lng_4[1], lat_4[1]),
          QgsRectangle(lng_4[1], lat_4[0], lng_4[2], lat_4[1]),
          QgsRectangle(lng_4[0], lat_4[1], lng_4[1], lat_4[2]),
          QgsRectangle(lng_4[1], lat_4[1], lng_4[2], lat_4[2])
        ]
        for rect in rects:
          cmptRadius(results, QgsReferencedRectangle(rect, target_rect.crs()))
        
    results = []
    cmptRadius(results, self.FETCH_AREA)
    
    for res in results:
      params_estat = self.ESTAT_API_PARAMS
      params_estat.update(res)
      self.WEBFETCH_ARGS["REQUEST_ARGS"].append(
        {
          "method": "GET",
          "url": self.parameterAsString(parameters, "WEBFETCH_URL", context) + f"?{urllib.parse.urlencode(params_estat)}"
        }
      )
    

    if self.WEBFETCH_ARGS["CRS"] is not None:
        self.WEBFETCH_ARGS["SET"] = True
  
  # fetch features from web
  def fetchFeaturesFromWeb(self, parameters: dict, context: QgsProcessingContext, feedback: QgsProcessingFeedback) -> None:
    self.downloadFilesConcurrently(parameters, context, feedback)
    
    # then merge the downloaded files
    init_string = self.WEBFETCH_ARGS["GEOM_TYPE"] + "?crs=" + self.WEBFETCH_ARGS["CRS"].authid() + "&index=yes&field=population:integer"
    vec_layer = QgsVectorLayer(init_string,baseName = "layer_from_tile", providerLib = "memory")
    vec_pr = vec_layer.dataProvider()
    
    # if there are no files
    if len(self.WEBFETCH_ARGS["DOWNLOADED_FILE"]) == 0:
      feedback.reportError(self.tr("No population results were obtained!"))
      raise Exception(self.tr("No population results were obtained!"))
    
    
    for file in self.WEBFETCH_ARGS["DOWNLOADED_FILE"]:
      with open(file, 'r') as f:
        body = json.load(f)
      if body.get("GET_SUMMARY", {}).get("DATASET_INF") != None:
        # table_inf = body["GET_SUMMARY"]["DATASET_INF"][0]["TABLE_INF"]
        for table in body["GET_SUMMARY"]["DATASET_INF"][0]["TABLE_INF"]:
          if all([table.get(k) == v for k, v in self.ESTAT_WHAT.items()]):
            lnglats = table["CLASS_INF"]
            results = table["DATA_INF"]
        
    #     for stat_dict in results:
    #       lnglat = self.WEBFETCH_ARGS["MESH"].get(stat_dict["@area"][:4], {}).get(stat_dict["@area"])
    #       if lnglat is not None:
    #         ft = QgsFeature(vec_layer.fields())
    #         ft.setGeometry(QgsPoint(lnglat.get("LONG"), lnglat.get("LAT")))
    #         ft["population"] = stat_dict["$"]
    #         vec_pr.addFeatures([ft])
    
    # self.FETCH_FEATURE = processing.run(
    #   "gdal:rasterize",
    #   {
    #     "INPUT": vec_layer,
    #     "FIELD": "population",
    #     "UNITS": 1,
    #     "WIDTH": 11.25 / 3600.0,
    #     "HEIGHT": 7.5 / 3600.0,
    #     "DATA_TYPE": 2,
    #     "INIT": 0,
    #     "OUTPUT": self.parameterAsOutputLayer(parameters, "OUTPUT", context)
    #   }
    # )["OUTPUT"]
      
  def initAlgorithm(self, config):    
    self.initUsingCanvas()
    self.initParameters()
  
  
  def processAlgorithm(self, parameters, context, feedback):  
    
    self.setFetchArea(parameters,context,feedback,self.parameterAsCrs(parameters, "FETCH_CRS", context))
    self.setWebFetchArgs(parameters, context, feedback)
    self.fetchFeaturesFromWeb(parameters, context, feedback)
    self.setCurrentProcess(use_caller = False)
    self.registerPostProcessor(
      self.FETCH_FEATURE, 
      hrPostProcessor(history = [self.CURRENT_PROCESS], color_args = {"coloring": "single_band_pseudo_color", "theme": "Purples", "opacity": 0.5})      
    )
    PostProcessors[self.FETCH_FEATURE].setMinValToZero()
    return {"OUTPUT":self.FETCH_FEATURE}
    
  def displayName(self):
    return self.tr("Population (Ja)")

  def group(self):
    return self.tr('Fetch geometries (Ja)')

  def groupId(self):
    return 'fetchgeomja'

  def createInstance(self):
    return fetchpopja()
