import typing
from qgis.PyQt.QtCore import (
  QVariant
  )
from qgis.core import (
  QgsProcessingFeedback,
  QgsProcessingContext,
  QgsField,
  QgsFields,
  QgsFeature,
  QgsVectorLayer
  )


import datetime

from ..algabstract import algabstract
from ..algpostprocessor import hrPostProcessor

# this is the abstract class, which will be inherited by the algorithms to initialize geometries
class initabstract(algabstract):
    
  def __init__(self):
    super().__init__()
    self.PK = 1 # the primary key, which is used to identify the features
    self.INPUT_LAYER = None  # temporal layer to contain the input features
    
    # the dictionaries of fields to initialize geometries are:
    # FIELDS_INIT: the fields of the output layer, which is used to initialize the output layer
    # FIELDS_REQUIRED_COMMON: the fields that are required to be contained in the output layer
    # FIELDS_REQUIRED: required fields though the overwrite settings are controlled by the user
    #
    # PROCEDURE:
    # first, FIELDS_REQUIRED_COMMON is loaded to FIELDS_INIT
    # second, existing fields are added to FIELDS_INIT (FIELDS_REQUIRED_COMMON is not overwritten)
    # third, FIELDS_REQUIRED is added to FIELDS_INIT
    #  where, if a field is duplicated and the overwrite flag is True, FIELDS_REQUIRED overwrites existing FIELD_INIT
    
    self.FIELDS_INIT = {
      # name of field: {
      #   "NAME": name of the field
      #   "TYPE": data type of the field, specified by QVariant (QVariant.Int, QVariant.String, QVariant.Double)
      #   "SOURCE": source of the field, "EXISTING", "APPEND_EXISTING" or "NEW"
      #   "DEFAULT_VALUE": default value, used if the SOURCE is NOT "EXISTING". geometry functions of $x, $y and $z are also acceptable.
      # }
    }
    self.FIELDS_REQUIRED_COMMON = {
      # name of field: {"TYPE": data type of the field, "DEFAULT_VALUE": default value}
      "PK":      {"NAME": "PK","TYPE": QVariant.Int, "SOURCE": "NEW","DEFAULT_VALUE": None},
      "HISTORY": {"NAME": "HISTORY","TYPE": QVariant.String, "SOURCE": "APPEND_EXISTING","DEFAULT_VALUE": None}
      }
    self.FIELDS_REQUIRED = {
      # name of field: {
      #   "TYPE": data type of the field, specified by QVariant (QVariant.Int, QVariant.String, QVariant.Double)
      #   "SOURCE": source of the field, "APPEND_EXISTING" or "NEW"
      #   "DEFAULT_VALUE": default value, used if the SOURCE is NOT "EXISTING". geometry functions of $x, $y and $z are also acceptable.
      # }
    } 
  
  def parseFieldDict(self, fld_dict: dict) -> QgsFields:
    fields = QgsFields()
    for fld_name, fld_info in fld_dict.items():
      fields.append(QgsField(fld_name, fld_info["TYPE"]))
    return fields

  # set the dictionaries of fields to initialize geometries
  def setFieldsDict(self, parameters: dict, context: QgsProcessingContext, feedback: QgsProcessingFeedback, existing_fields:QgsFields = QgsFields(), overwrite: bool = False) -> None:
    # initialize the list of the dictionary of fields to initialize
    self.FIELDS_INIT = self.FIELDS_REQUIRED_COMMON
    
    # add existing fields to the FIELDS_INIT    
    for existing_fld in existing_fields:
      if existing_fld.name().lower() not in map(lambda x: x.lower(), self.FIELDS_INIT.keys()):
        self.FIELDS_INIT[existing_fld.name()] = {
          "NAME": existing_fld.name(), "TYPE": existing_fld.type(), "SOURCE": "EXISTING", "DEFAULT_VALUE": None
        }
    
    # add the additional fields to the FIELDS_INIT    
    for fld_name, fld_info in self.FIELDS_REQUIRED.items():
      if fld_name.lower() in map(lambda x: x.lower(), self.FIELDS_INIT.keys()):
        if overwrite:
          self.FIELDS_INIT[fld_name] = fld_info
          self.FIELDS_INIT[fld_name]["NAME"] = fld_name
          self.FIELDS_INIT[fld_name]["SOURCE"] = "NEW (OVERWRITE)"
      else:
        self.FIELDS_INIT[fld_name] = fld_info
        self.FIELDS_INIT[fld_name]["NAME"] = fld_name
        self.FIELDS_INIT[fld_name]["SOURCE"] = "NEW"
    
    # feedback the results
    feedback.pushInfo(self.tr("The following fields are initialized: "))
    feedback.pushInfo(str({k:{"TYPE": v["TYPE"], "SOURCE": v["SOURCE"]} for k,v in self.FIELDS_INIT.items()}))
    
  
  # reset the counter
  def resetPk(self):
    self.PK = 0
  
  def parseDefaultValue(self, fld_info: dict, ft: QgsFeature):
    if fld_info["NAME"] == "PK":
      self.PK += 1
      return self.PK
    elif fld_info["NAME"] == "HISTORY":
      return self.CURRENT_PROCESS
    else:
      try:
        if fld_info["DEFAULT_VALUE"] == "$x":
          return ft.geometry().get().x()
        elif fld_info["DEFAULT_VALUE"] == "$y":
          return ft.geometry().get().y()
        elif fld_info["DEFAULT_VALUE"] == "$z":
          return ft.geometry().get().z()
      except:
        raise Exception(self.tr("The following geometry function fails: ") + fld_info["DEFAULT_VALUE"])
      return fld_info["DEFAULT_VALUE"]
      
  
  # add features to the sink, based on the fields dictionaries  
  def addFeaturesToSink(self, sink, dest_id, existing_layer: QgsVectorLayer = None):
    
    if existing_layer is None:
      return
    
    # set the current process
    self.setCurrentProcess(use_caller = True)
    self.registerPostProcessor(dest_id, hrPostProcessor(history = [self.CURRENT_PROCESS]))
    
    # add features
    for ft in existing_layer.getFeatures():
      # initialize the feature
      new_ft = QgsFeature(self.parseFieldDict(self.FIELDS_INIT))
      new_ft.setGeometry(ft.geometry())
      
      for fld_name, fld_info in self.FIELDS_INIT.items():
        # the fields contained in the input existing_layer are copied
        if fld_info.get("SOURCE") == "EXISTING" and fld_name in ft.fields().names():
          new_ft[fld_name] = ft[fld_name]
        
        else:
          def_val = self.parseDefaultValue(fld_info, ft)
          
          if fld_info.get("SOURCE") == "APPEND_EXISTING" and fld_name in ft.fields().names() and isinstance(ft[fld_name], str) and len(ft[fld_name]) > 0:
              new_ft[fld_name] = ft[fld_name] + "; " + def_val
          else:
            new_ft[fld_name] = def_val
            
      sink.addFeatures([new_ft])
          