
from qgis.PyQt.QtCore import (QCoreApplication, QT_TRANSLATE_NOOP, QVariant)
from qgis.core import (
  QgsVectorLayer,
  QgsProcessingContext,
  QgsReferencedRectangle,
  QgsWkbTypes,
  QgsField,
  QgsFields,
  QgsFeature,
  QgsGeometry,
  QgsProcessingFeedback
  )

from qgis import processing

import os
import re
import datetime

from ..algabstract import algabstract
from ..algpostprocessor import hrPostProcessor
from ..hr_config import PostProcessors

class noiseabstract(algabstract):
  
  BLDG_LEVEL_ARGS = {
    "BUILDING_BID": "PK", # primary key of the buildings
    "RECEIVER_BID": "BUILD_PK", # primary key of the buildings, written in receiver layer
    "RECEIVER_RID": "PK", # primary key of the receivers
    "LEVEL_RID": "IDRECEIVER", #primary key of the receivers, written in level layer    
    "LEVEL_ASSIGN": ["LAEQ"] # which attribute(s) to be assined
  }
  
  ISOSURFACE_ARGS = {
    "LEVEL_RID": "IDRECEIVER" #primary key of the receivers, written in level layer        
  }
    
  PROC_RESULTS = {}
  

  def initNoiseModellingPath(self, paths={}, feedback:QgsProcessingFeedback = None):
    paths.update(
      {
        "TRIANGLE": os.path.join("%nmtmp%", "TRIANGLES.geojson"),
        "LEVEL_RESULTS":  {
          "LDAY":     os.path.join("%nmtmp%", "LDAY_GEOM.geojson"),
          "LEVENING": os.path.join("%nmtmp%", "LEVENING_GEOM.geojson"),
          "LNIGHT":   os.path.join("%nmtmp%", "LNIGHT_GEOM.geojson"),
          "LDEN":     os.path.join("%nmtmp%", "LDEN_GEOM.geojson"),
        },
        "BUILDING_RESULTS": {
          "BUILDING_WITH_LEVEL": os.path.join("%nmtmp%", "BUILDING_WITH_LEVEL.geojson")
        }
      }
    )
    super().initNoiseModellingPath(paths, feedback = feedback)
    
  # output a polygon (sink and output the `dest_id`) 
  # that stores arguments of the calculation
  def outputWpsArgs(self, parameters:dict, context:QgsProcessingContext, extent_rec: QgsReferencedRectangle) -> str:  
    args_fields = QgsFields()
    args_fields.append(QgsField("NoiseModelling version", QVariant.String))
    args_fields.append(QgsField("H-RISK version", QVariant.String))
    
    for key, value in self.NOISEMODELLING["WPS_ARGS"].items():
      if key == "w":
        key = "workingDirectory"
      elif key == "s":
        key = "scriptPath"
        
      if type(value) in [int, bool]:
        args_fields.append(QgsField(key, QVariant.Int))
      elif type(value) in [float]:
        args_fields.append(QgsField(key, QVariant.Double))
      elif type(value) in [str]:
        args_fields.append(QgsField(key, QVariant.String))
        
    (sink, dest_id) = self.parameterAsSink(
      parameters, "WPS_ARGS", context, args_fields, QgsWkbTypes.Polygon, extent_rec.crs()
      )
    
    ft = QgsFeature(args_fields)
    ft.setGeometry(QgsGeometry.fromRect(extent_rec))
    ft["NoiseModelling version"] = re.search(r"NoiseModelling v (\S+)", self.CURRENT_PROCESS).group(1)
    ft["H-RISK version"] = re.search(r"H-RISK v (\S+)", self.CURRENT_PROCESS).group(1)
    
    for key, value in self.NOISEMODELLING["WPS_ARGS"].items():
      if key == "w":
        key = "workingDirectory"
      elif key == "s":
        key = "scriptPath"
        
      ft[key] = value
      
    sink.addFeature(ft)
    self.setCurrentProcess(use_caller=True)
    self.registerPostProcessor(dest_id, hrPostProcessor(history = [self.CURRENT_PROCESS]))
    
    return dest_id
  
  # create sound-level assigned buildings
  def cmptBuildingLevel(
    self, parameters: dict, context: QgsProcessingContext, feedback: QgsProcessingFeedback, bldg_layer: QgsVectorLayer, rcv_layer: QgsVectorLayer,
    level_subtraction: float = 3.0
    ) -> None:
    
    for noise_idx, file_path in self.NOISEMODELLING["LEVEL_RESULTS"].items():          
      # overwrite building layer
      estimate_level_args = {
          "BUILDING": bldg_layer,
          "RECEIVER": rcv_layer,
          "LEVEL": QgsVectorLayer(file_path, noise_idx),
          "LEVEL_PREFIX": noise_idx + "_",
          "SUBTRACTION_TO_CANCEL_REFLECTION": level_subtraction,
          "OVERWRITE": True,
          "OUTPUT": "TEMPORARY_OUTPUT"
          }
      estimate_level_args.update(self.BLDG_LEVEL_ARGS)
      
      bldg_layer = processing.run(
        "hrisk:estimatelevelofbuilding",
        estimate_level_args
      )["OUTPUT"]
    
    # estimate the health risk
    try:
      bldg_layer = processing.run(
        "hrisk:estimateriskofbuilding",
        {
          "BUILDING": bldg_layer,
          "LDEN": "LDEN_LAEQ_maximum",
          "LNIGHT": "LNIGHT_LAEQ_maximum",
          "POP": "pop",
          "OVERWRITE": True,
          "OUTPUT": "TEMPORARY_OUTPUT"
        }
      )["OUTPUT"]
    except Exception as e:
      raise Exception(self.tr("Sound levels are not assigned for buildings."))
    
    # save the result
    self.saveVectorLayer(bldg_layer, self.NOISEMODELLING["BUILDING_RESULTS"]["BUILDING_WITH_LEVEL"] )

  def setPostProcessors(self):
    group_name = "results_" + datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    for layer_name, dest_id in self.PROC_RESULTS.items():
      PostProcessors[dest_id].setGroup(group_name)
      if layer_name in ["WPS_ARGS"]:
        PostProcessors[dest_id].setColorArgs(coloring = "bbox")
      elif layer_name in ["LDAY","LEVENING", "LNIGHT", "LDEN"]:
        PostProcessors[dest_id].setColorArgs(coloring = "graduated_symbol", attribute = "LAEQ", theme = "END")
      elif layer_name == "BUILDING_WITH_LEVEL":
        PostProcessors[dest_id].setColorArgs(coloring = "graduated_symbol", attribute = "LDEN_LAEQ_maximum", theme = "END")
      elif layer_name == "CONTOURING_NOISE_MAP":
        PostProcessors[dest_id].setGroup(None)
        PostProcessors[dest_id].setColorArgs(coloring = "categorized_symbol", attribute = "ISOLABEL", theme = "END")