from qgis.PyQt.QtGui import QColor
from qgis.core import (
  QgsProcessingLayerPostProcessorInterface,
  QgsSingleBandPseudoColorRenderer,
  QgsRasterShader,
  QgsColorRampShader,
  QgsProcessingLayerPostProcessorInterface,
  QgsRasterBandStats,
  QgsProject,
  QgsFillSymbol,
  QgsGraduatedSymbolRenderer,
  QgsRendererRange,
  QgsClassificationRange,
  QgsMarkerSymbol,
  QgsFillSymbol,
  QgsLineSymbol,
  QgsWkbTypes,
  QgsCategorizedSymbolRenderer,
  QgsRendererCategory
  )

import re

class hrPostProcessor (QgsProcessingLayerPostProcessorInterface):
  
  # class variable
  COLOR_THEMES:dict = {
    # colors were generated using http://colorbrewer2.org/, with 6 classes
    "Greys": {
      "value_map": {
        "min": {"value": None, "color": QColor("#f7f7f7")},
        "max": {"value": None, "color": QColor("#252525")}
      }
    },
    "Greens": {
      "value_map": {
        "min": {"value": None, "color": QColor("#edf8e9")},
        "max": {"value": None, "color": QColor("#006d2c")}
      }
    },
    "Blues": {
      "value_map": {
        "min": {"value": None, "color": QColor("#eff3ff")},
        "max": {"value": None, "color": QColor("#08519c")}
      }
    },
    "Oranges": {
      "value_map": {
        "min": {"value": None, "color": QColor("#feedde")},
        "max": {"value": None, "color": QColor("#a63603")}
      }
    },
    "Purples": {
      "value_map": {
        "min": {"value": None, "color": QColor("#f2f0f7")},
        "max": {"value": None, "color": QColor("#54278f")}
      }
    },
    "Reds": {
      "value_map":{
        "min": {"value": None, "color": QColor("#fee5d9")},
        "max": {"value": None, "color": QColor("#a50f15")}
      }
    },
    "END": {
      "range_map":{
        "< 35 dB":    {"lower": -999, "upper":  35, "color": QColor(255, 255, 255)},
        "35 - 40 dB": {"lower":   35, "upper":  40, "color": QColor(160, 186, 191)},
        "40 - 45 dB": {"lower":   40, "upper":  45, "color": QColor(184, 214, 209)},
        "45 - 50 dB": {"lower":   45, "upper":  50, "color": QColor(206, 228, 204)},
        "50 - 55 dB": {"lower":   50, "upper":  55, "color": QColor(226, 242, 191)},
        "55 - 60 dB": {"lower":   55, "upper":  60, "color": QColor(243, 198, 131)},
        "60 - 65 dB": {"lower":   60, "upper":  65, "color": QColor(232, 126,  77)},
        "65 - 70 dB": {"lower":   65, "upper":  70, "color": QColor(205,  70,  62)},
        "70 - 75 dB": {"lower":   70, "upper":  75, "color": QColor(161,  26,  77)},
        "75 - 80 dB": {"lower":   75, "upper":  80, "color": QColor(117,   8,  92)},
        "> 80 dB":    {"lower":   80, "upper": 999, "color": QColor(67,   10,  74)}
      }
    }
  }
  
  # initialize instance variables
  def __init__(self, history:list = [], color_args:dict = {"coloring": "none"}, group = None, visibility:bool = True):
    super().__init__()
    self.HISTORY = []
    self.COLOR_ARGS = {
      "coloring": "none",
      "theme": None,
      "attribute": "LAEQ",
      "band_idx" : 1,
      "fill_color": QColor(0,0,0,0),
      "border_color": QColor(0,0,0,255),
      "border_width": 0.5,
      "value_map" : {"min": {"value": None, "color": QColor("#f7f7f7")}},
      "range_map" : {"default": {"lower": -999, "upper": 999, "color": QColor(0,0,0,255)}},
      "set_clip" : True,
      "opacity" : 0.5
    }
    self.GROUP = None
    self.VISIBILITY = True
      
    self.setHistory(history)
    self.setColorArgs(**color_args)
    self.setGroup(group)
    self.setVisivility(visibility)

  # clone the instance
  def clone(self):
    kwargs = {
      "history": self.HISTORY,
      "color_args": self.COLOR_ARGS,
      "group": self.GROUP,
      "visibility": self.VISIBILITY
    }
    return hrPostProcessor(**kwargs)
  
  # group setter
  def setGroup(self, group):
    self.GROUP = group
  
  # visibility setter
  def setVisivility(self, visibility):
    self.VISIBILITY = visibility
  
  # color arguments setter
  def setColorArgs(self, **kwargs):
    if "theme" in kwargs.keys() and kwargs["theme"] in self.COLOR_THEMES:
      self.setColorArgs(**self.COLOR_THEMES[kwargs["theme"]])
      
    for k, v in kwargs.items():
      if k in self.COLOR_ARGS.keys():
        self.COLOR_ARGS[k] = v
      else:
        raise Exception("Unknown argument: " + str(k))
    
    if "opacity" in kwargs.keys() and kwargs["opacity"] >= 0.0 and kwargs["opacity"] <= 1.0:
      for arg in self.COLOR_ARGS.values():
        if isinstance(arg, dict):
          for val in arg.values():
            if isinstance(val, dict) and "color" in val.keys():
              val["color"].setAlphaF(kwargs["opacity"])
  
  # the minimum value of the color range is set to zero
  def setMinValToZero(self):
    self.COLOR_ARGS["value_map"]["min"]["value"] = 0.0
    
  # the minimum and maximum values of the color range are set to the minimum and maximum values of the layer
  def setValRangeUsingStats(self, layer, only_if_unset = True):
    stats = layer.dataProvider().bandStatistics(self.COLOR_ARGS["band_idx"], QgsRasterBandStats.All) 
    if only_if_unset:
      if self.COLOR_ARGS["value_map"]["min"]["value"] is None:
        self.COLOR_ARGS["value_map"]["min"]["value"] = stats.minimumValue
      if self.COLOR_ARGS["value_map"]["max"]["value"] is None:
        self.COLOR_ARGS["value_map"]["max"]["value"] = stats.maximumValue
    else:
      self.COLOR_ARGS["value_map"]["min"]["value"] = stats.minimumValue
      self.COLOR_ARGS["value_map"]["max"]["value"] = stats.maximumValue


  def setCategorizedColUsingRangeVals(self, layer):
    self.COLOR_ARGS["value_map"] = {}
    for ft in layer.getFeatures():
      if not ft[self.COLOR_ARGS["attribute"]] in self.COLOR_ARGS["value_map"].keys():
        nums_in_attr = [float(s) for s in re.findall(r"\d+\.*\d*", ft[self.COLOR_ARGS["attribute"]])]
        if re.match(r"<", ft[self.COLOR_ARGS["attribute"]]) != None:
          nums_in_attr.append(-999)
        nums_in_attr_mean = sum(nums_in_attr) / len(nums_in_attr)
        for color_map in self.COLOR_ARGS["range_map"].values():
          if nums_in_attr_mean >= color_map["lower"] and nums_in_attr_mean < color_map["upper"]:
            self.COLOR_ARGS["value_map"][ft[self.COLOR_ARGS["attribute"]]] = color_map

      

  # history setter
  def setHistory(self, history:list):
    self.HISTORY = history

  # history getter
  def getHistory(self):
    return self.HISTORY
  
  # history extender
  def extendHistory(self, history:list):
    self.HISTORY.extend(history)
  
  # Single band pseudo color renderer for a raster layer
  def setSingleBandPseudoColorRenderer(self, layer):    
    if layer.__class__.__name__ != "QgsRasterLayer":
      raise Exception("Single band pseudo color renderer is only applicable to a raster layer")
    if "min" not in self.COLOR_ARGS["value_map"].keys() or "max" not in self.COLOR_ARGS["value_map"].keys():
      self.setColorRangeUsingTheme("Greys")
      
    if self.COLOR_ARGS["value_map"]["min"]["value"] is None or self.COLOR_ARGS["value_map"]["max"]["value"] is None:
      self.setValRangeUsingStats(layer)
    
    cr_fun = QgsColorRampShader(
      minimumValue = self.COLOR_ARGS["value_map"]["min"]["value"],
      maximumValue = self.COLOR_ARGS["value_map"]["max"]["value"]
    )
    cr_fun.setColorRampItemList(
      [
        QgsColorRampShader.ColorRampItem(v["value"], v["color"]) for v in self.COLOR_ARGS["value_map"].values()
      ]
    )
    cr_fun.setClip(self.COLOR_ARGS["set_clip"])
    shader = QgsRasterShader()
    shader.setRasterShaderFunction(cr_fun)
    lyr_renderer = QgsSingleBandPseudoColorRenderer(layer.dataProvider(), self.COLOR_ARGS["band_idx"], shader)
    layer.setRenderer(lyr_renderer)
    
  
  # graduated symbol renderer for a vector layer
  def setGraduatedSymbolRenderer(self, layer):
    if layer.__class__.__name__ != "QgsVectorLayer":
      raise Exception("Graduated symbol renderer is only applicable to a vector layer")
    
    # change symboller according to the wkb-type
    symb_fun = None
    if layer.wkbType() in [QgsWkbTypes.Point, QgsWkbTypes.PointZ]:
      symb_fun = QgsMarkerSymbol.createSimple
    elif layer.wkbType() == QgsWkbTypes.Polygon:
      symb_fun = QgsFillSymbol.createSimple
    elif layer.wkbType() == QgsWkbTypes.LineString:
      symb_fun = QgsLineSymbol.createSimple
    else:
      raise Exception("Unsupported geometry type: " + str(layer.wkbType())) 
    
    # set renderer and symbols
    try:
      renderer = QgsGraduatedSymbolRenderer(self.COLOR_ARGS["attribute"])
      for key, value in self.COLOR_ARGS["range_map"].items():
        renderer.addClassRange(
          QgsRendererRange(
            QgsClassificationRange(key, value.get("lower"), value.get("upper")), 
            symb_fun({"color":value.get("color")})
          )
        )
      # apply renderer
      layer.setRenderer(renderer)
    except:
      raise Exception("Error in setting the graduated symbol renderer")
      
      
  # use a single symbol renderer for a vector layer
  def setFillSymbolRenderer(self, layer):
    if layer.__class__.__name__ != "QgsVectorLayer":
      raise Exception("Single symbol renderer is only applicable to a vector layer")
    symbol = QgsFillSymbol.createSimple({"color":self.COLOR_ARGS["fill_color"], "color_border":self.COLOR_ARGS["border_color"], "width_border":self.COLOR_ARGS["border_width"]})
    layer.renderer().setSymbol(symbol)
    layer.setOpacity(self.COLOR_ARGS["opacity"])
  
  def setCategorizedSymbolRenderer(self, layer):
      # set renderer
    renderer = QgsCategorizedSymbolRenderer("ISOLABEL")
                
    # set symbols
    for key, value in self.COLOR_ARGS["value_map"].items():
      renderer.addCategory(
        QgsRendererCategory(
          key, QgsFillSymbol.createSimple({"color":value.get("color")}), key + " dB"
        )
      )
    # apply renderer
    layer.setRenderer(renderer)
  
  # Post processing MAIN
  def postProcessLayer(self, layer, context, feedback):
    
    # set history items
    layer_meta = layer.metadata()
    if len(self.HISTORY) > 0:
      for history in self.HISTORY:
        layer_meta.addHistoryItem(history)
    layer.setMetadata(layer_meta)
    
    # set renderer
    if self.COLOR_ARGS["coloring"] != "none":
      if self.COLOR_ARGS["coloring"] == "single_band_pseudo_color":
        self.setSingleBandPseudoColorRenderer(layer)
      elif self.COLOR_ARGS["coloring"] == "bbox":
        self.setFillSymbolRenderer(layer)      
      elif self.COLOR_ARGS["coloring"] == "graduated_symbol":
        self.setGraduatedSymbolRenderer(layer)
      elif self.COLOR_ARGS["coloring"] == "categorized_symbol":
        self.setCategorizedColUsingRangeVals(layer)
        self.setCategorizedSymbolRenderer(layer)
      
      layer.triggerRepaint()        
    
    # set group
    if self.GROUP is not None:      
      root = QgsProject.instance().layerTreeRoot()
      
      grp = root.findGroup(self.GROUP)
      if grp is None:
        root.insertGroup(0, self.GROUP)
        grp = root.findGroup(self.GROUP)
        
      vl = root.findLayer(layer.id())
      vl.setItemVisibilityChecked(self.VISIBILITY)
      vl_clone = vl.clone()
      parent = vl.parent()
      grp.insertChildNode(0, vl_clone)
      parent.removeChildNode(vl)
      
